Require Import vgtac cola.

Open Scope cola_scope.
Set Implicit Arguments.

Definition img X Y (f: X -> Y) (P: X -> Prop) : Y -> Prop := 
  fun y => exists x, P x /\ y = f x.
Hint Unfold img.

Lemma img_compose: forall {X Y Z} {f: Y -> Z} {g: X -> Y} {P: X -> Prop},
  img f (img g P) == img (f <*> g) P.
Proof. unfold img; split; ii; des; subst; eauto. Qed.


(** Basic Mendler-style monotonizations and fixed points.
*)

Section basic_mendler_upper.

  Variable C : cola.

  Definition uppermon (g : C -> C) (c : C) : C :=
    cola_lub (img g (fun c' => c' <= c)).

  Lemma uppermon_mon : forall g, monotone (uppermon g).
  Proof.
    i; unfold monotone; i; eapply cola_lub_least; i; destruct H, H.
    eapply cola_lub_upper; exists x0; eauto using cola_refl.
    split; try eapply cola_trans; eauto.
  Qed.

  Lemma uppermon_galois : forall f g (MON : monotone f),
                            uppermon g <= f <-> g <= f.
  Proof.
    split; i; intro.
    - specialize (H x); rewrite <- H; apply cola_lub_upper; eauto.
    - apply cola_lub_least; i; rdes H0; subst; rewrite <- (MON _ _ H0); eauto.
  Qed.

  Corollary uppermon_ge : forall g, g <= uppermon g.
  Proof.
    i; eapply uppermon_galois; eauto using uppermon_mon.
  Qed.

  Lemma uppermon_eq: forall g (MON : monotone g), g == (uppermon g).
  Proof.
    split; ii.
    - apply uppermon_ge.
    - eapply cola_lub_least; i; rdes H; subst; eauto.
  Qed.

  Theorem fixp_uppermon : forall g (MON : monotone g),
                            fixp g == fixp (uppermon g).
  Proof.
    split; eapply fixp_mon; eauto using uppermon_mon; apply uppermon_eq; eauto.
  Qed.

  Theorem cofixp_uppermon : forall g (MON : monotone g),
                              cofixp g == cofixp (uppermon g).
  Proof.
    split; eapply cofixp_mon; eauto using uppermon_mon; apply uppermon_eq; eauto.
  Qed.

End basic_mendler_upper.


Section basic_mendler_lower.

  Variable C : cola.

  Definition lowermon (g : C -> C) (c : C) : C :=
    cola_glb (img g (fun c' => c <= c')).

  Lemma lowermon_mon : forall g, monotone (lowermon g).
  Proof.
    ii; eapply (@uppermon_mon (dual_cola C)); eauto.
  Qed.

  Lemma lowermon_galois : forall f (MON : monotone f) g,
                            f <= lowermon g <-> f <= g.
  Proof.
    ii; eapply (@uppermon_galois (dual_cola C)); ii; apply MON; eauto.
  Qed.

  Corollary lowermon_le : forall g, lowermon g <= g.
  Proof.
    ii; eapply (@uppermon_ge (dual_cola C)).
  Qed.

  Theorem lowermon_eq: forall g (MON : monotone g), g == lowermon g.
  Proof.
    ii; symmetry; eapply (@uppermon_eq (dual_cola C)); ii; apply MON; eauto.
  Qed.

  Theorem fixp_lowermon : forall g (MON : monotone g),
                            fixp g == fixp (lowermon g).
  Proof.
    ii; symmetry; eapply (@cofixp_uppermon (dual_cola C)); ii; apply MON; eauto.
  Qed.

  Theorem cofixp_lowermon : forall g (MON : monotone g),
                              cofixp g == cofixp (lowermon g).
  Proof.
    ii; symmetry; eapply (@fixp_uppermon (dual_cola C)); ii; apply MON; eauto.
  Qed.

End basic_mendler_lower.


(** Generalized Mendler-style monotonizations and fixed points.
*)

Section generalized_mendler_upper.

  Variable D C : cola.
  Variable inj : D -> C.
  Hypothesis inj_embed : forall d d', d <= d' <-> inj d <= inj d'.
  Hypothesis inj_preserves_lub : forall (S : D -> Prop),
                                   inj (cola_lub S) <= cola_lub (img inj S).

  Definition uppermon_i (g : D -> D) (c : C) : C :=
    cola_lub (img (inj <*> g) (fun d => inj d <= c)).

  Definition upperright (f: C -> C) (d : D) : D := 
    cola_lub (fun d' => inj d' <= f (inj d)).

  Lemma inj_preserves_lub_r: forall (S: D -> Prop),
                               cola_lub (img inj S) <= inj (cola_lub S).
  Proof.
    by i; eapply cola_lub_least; i; rdes H; subst; eapply inj_embed, cola_lub_upper.
  Qed.

  Lemma uppermon_i_mon : forall g, monotone (uppermon_i g).
  Proof.
    i; unfold monotone; i; eapply cola_lub_least; i; destruct H, H.
    eapply cola_lub_upper; exists x0; eauto using cola_refl.
    split; try eapply cola_trans; eauto.
  Qed.

  Lemma uppermon_i_galois : forall f g (MON : monotone f),
                              uppermon_i g <= f <-> g <= upperright f.
  Proof.
    split; i; intro.
    - assert (forall x y, inj x <= y -> inj (g x) <= f y).
      + by i; rr in H; rewrite <- H; eauto using cola_lub_upper.
      + by eapply cola_lub_upper; eauto.
    - assert (forall x y, inj x <= y -> inj (g x) <= f y).
      + i; specialize (H x0); eapply inj_embed in H.
        unfold upperright, img in *.
        rewrite H; erewrite inj_preserves_lub.
        by eapply cola_lub_least; i; des; subst; rewrite H1; eauto.
      + by eapply cola_lub_least; unfold img; i; des; subst; eauto.
  Qed.

  Theorem fixp_uppermon_i : forall (g : D -> D) (MON: monotone g),
                              inj (fixp g) == fixp (uppermon_i g).
  Proof.
    r; splits.
    - rewrite <-(@le_fixp C); eauto using uppermon_i_mon.
      unfold uppermon_i at 1; rewrite <-(proj1 img_compose), <-inj_preserves_lub.
      apply inj_embed, fixp_ind, cola_lub_upper.
      unfold img at 1; esplits; eauto. 
      rewrite inj_preserves_lub, (proj1 img_compose), <-le_fixp; eauto using uppermon_i_mon.
    - apply fixp_ind; unfold uppermon_i; rewrite (proj2 img_compose), inj_preserves_lub_r.
      eapply inj_embed, cola_lub_least; i; rdes H; subst. 
      rewrite <-le_fixp; eauto; apply MON, inj_embed; eauto.
  Qed.

  Theorem cofixp_uppermon_i : forall (g : D -> D) (MON: monotone g),
                                inj (cofixp g) == cofixp (uppermon_i g).
  Proof.
    r; splits.
    - apply cofixp_coind; unfold uppermon_i; rewrite <-(proj1 img_compose),<-inj_preserves_lub.
      eapply inj_embed; rewrite cofixp_le, cola_lub_upper; eauto; inst; eauto.
    - rewrite cofixp_le; eauto using uppermon_i_mon.
      unfold uppermon_i at 1; rewrite (proj2 img_compose), inj_preserves_lub_r.
      apply inj_embed, cofixp_coind, cola_lub_least.
      i; rdes H; subst; apply MON, inj_embed. 
      rewrite H, <-inj_preserves_lub_r, <-(proj2 img_compose), cofixp_le; 
        eauto using uppermon_i_mon.
  Qed.

End generalized_mendler_upper.

Section generalized_mendler_lower.

  Variable D C : cola.
  Variable inj : D -> C.
  Hypothesis inj_embed : forall d d', d <= d' <-> inj d <= inj d'.
  Hypothesis inj_preserves_glb : forall (S : D -> Prop),
                                   cola_glb (img inj S) <= inj (cola_glb S).

  Definition lowermon_i (g : D -> D) (c : C) : C :=
    cola_glb (img (inj <*> g) (fun d => c <= inj d)).

  Definition lowerleft (f: C -> C) (d : D) : D := 
    cola_glb (fun d' => f (inj d) <= inj d').

  Lemma lowermon_i_mon : forall g, monotone (lowermon_i g).
  Proof. ii; eapply (@uppermon_i_mon (dual_cola D) (dual_cola C)); eauto. Qed.

  Lemma lowermon_i_galois : forall f g (MON : monotone f),
                              lowerleft f <= g <-> f <= lowermon_i g.
  Proof. 
    i; symmetry; eapply (@uppermon_i_galois (dual_cola D) (dual_cola C)); eauto.
    - ii; apply inj_embed.
    - ii; apply MON; eauto.
  Qed.

  Theorem fixp_lowermon_i : forall (g : D -> D) (MON: monotone g),
                              inj (fixp g) == fixp (lowermon_i g).
  Proof. 
    ii; symmetry; eapply (@cofixp_uppermon_i (dual_cola D) (dual_cola C)); eauto.
    - ii; apply inj_embed.
    - ii; apply MON; eauto.
  Qed.

  Theorem cofixp_lowermon_i : forall (g : D -> D) (MON: monotone g),
                                inj (cofixp g) == cofixp (lowermon_i g).
  Proof.
    ii; symmetry; eapply (@fixp_uppermon_i (dual_cola D) (dual_cola C)); eauto.
    - ii; apply inj_embed.
    - ii; apply MON; eauto.
  Qed.

End generalized_mendler_lower.
